#
# Collective Knowledge (classify image using various models such as DNN)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
hextra='<i><center>\n'
hextra+=' [ <a href="http://cKnowledge.org/ai">Community-driven AI R&D powered by CK</a> ], '
hextra+=' [ <a href="https://github.com/dividiti/ck-caffe">CK-Caffe</a> ], '
hextra+=' [ <a href="https://github.com/ctuning/ck-caffe2">CK-Caffe2</a> ], '
hextra+=' [ <a href="https://github.com/ctuning/ck-tensorflow">CK-TensorFlow</a> ], '
hextra+=' [ <a href="https://en.wikipedia.org/wiki/Collective_Knowledge_(software)">Wikipedia</a>, \n'
hextra+='<a href="https://www.researchgate.net/publication/304010295_Collective_Knowledge_Towards_RD_Sustainability">paper 1</a>, \n'
hextra+='<a href="https://arxiv.org/abs/1506.06256">Paper 2</a>, \n'
hextra+='<a href="https://www.youtube.com/watch?v=Q94yWxXUMP0">YouTube CK intro</a> ] \n'
hextra+='</center></i>\n'
hextra+='<br>\n'

hextra+='Optimizations results are continuously shared by volunteers across diverse programs, data sets and platforms: '
hextra+='<a href="http://cknowledge.org/repo/web.php?native_action=show&native_module_uoa=program.optimization&scenario=4dcf435bb0d92fa6">GCC</a> , \n'
hextra+='<a href="http://cknowledge.org/repo/web.php?native_action=show&native_module_uoa=program.optimization&scenario=1eb2f50d4620903e">LLVM</a> \n'

dlabels='collective_training_set'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# show dashboard

def show(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import shutil
    import time
    import copy

    form_name='milepost_web_form'

    sh=i.get('skip_html','')

    ww=False
    if i.get('widget','')=='yes': ww=True

    # State reset
    dar=False
    if 'dnn_action_reset' in i: 
       dar=True

    # Start HTML
    h=''
    st=''

    h+='<center>\n'
    h+='\n\n<script language="JavaScript">function copyToClipboard (text) {window.prompt ("Copy to clipboard: Ctrl+C, Enter", text);}</script>\n\n' 

    if not ww:
       h+=hextra

    if ww:
       url0=i.get('prepared_url0','')
       url1=i.get('prepared_url1','')
       form_name=i.get('prepared_form_name','')
    else:
       # Check host URL prefix and default module/action
       rx=ck.access({'action':'form_url_prefix',
                     'module_uoa':'wfe',
                     'host':i.get('host',''), 
                     'port':i.get('port',''), 
                     'template':i.get('template','')})
       if rx['return']>0: return rx
       url0=rx['url']
       template=rx['template']

       url=url0
       action=i.get('action','')
       muoa=i.get('module_uoa','')

       url+='action=index&module_uoa=wfe&native_action='+action+'&'+'native_module_uoa='+muoa
       url1=url

       # Start form
       r=ck.access({'action':'start_form',
                    'module_uoa':cfg['module_deps']['wfe'],
                    'url':url1,
                    'name':form_name})
       if r['return']>0: return r
       h+=r['html']

    url=url0
    onchange='document.'+form_name+'.submit();'

    # Header
    if not ww:
       h+='<h2>Unified CK AI JSON API to classify images using different DNN engines (Caffe, TensorFlow, etc) while monitoring/optimizing performance</h2>\n'

    # Select engine
    dt=[
        {'name':'Caffe', 'value':'caffe'},
        {'name':'Caffe2','value':'caffe2'},
        {'name':'TensorFlow', 'value':'tensorflow'}
       ]

    engine=i.get('dnn_engine','')
    if engine=='': engine='caffe'

    ii={'action':'create_selector',
        'module_uoa':cfg['module_deps']['wfe'],
        'data':dt,
        'name':'dnn_engine',
        'onchange':onchange, 
        'skip_sort':'yes',
        'selected_value':engine}
    r=ck.access(ii)
    if r['return']>0: return r
    x=r['html']

    h+='DNN engine: '+x+'<br><br>\n'

    # Search env to check if has installed
    r=ck.access({'action':'search',
                 'module_uoa':cfg['module_deps']['env'],
                 'tags':'lib,'+engine})
    if r['return']>0: return r
    el=r['lst']

    warning=''
    prediction=''

    if len(el)==0:
       h+='<i>DNN engine is not installed - contact <a href="mailto:admin@cKnowledge.org">admin</a> to install this engine and related models in the CK AI cloud</i><br><br>'
       warning='DNN engine is not installed'
    else:
       h+='<i>'+str(len(el))+' engine(s) installed (different optimizations and platforms)</i><br>'

       # Search models to check if has installed
       tx=engine+'model'
       if engine=='caffe2':
          tx='caffemodel2'

       r=ck.access({'action':'search',
                    'module_uoa':cfg['module_deps']['env'],
                    'tags':tx})
       if r['return']>0: return r
       models=r['lst']

       if len(models)==0:
          h+='<i>DNN models for this engine are not installed - contact <a href="mailto:admin@cKnowledge.org">admin</a> to install more models in the CK AI cloud</i><br><br>'
          warning='DNN models for this engine are not installed'
       else:
          h+='<i>'+str(len(models))+' model(s) installed (different topology and parameters)</i><br><br>'

          h+='<i>Optimization statistics shared by the community:\n'
          h+='[ <a href="http://cknowledge.org/repo/web.php?native_action=show&native_module_uoa=program.optimization&scenario=1eb2f50d4620903e">desktops and servers</a> ], \n'
          h+='[ <a href="http://cknowledge.org/repo/web.php?native_action=show&native_module_uoa=program.optimization&scenario=4dcf435bb0d92fa6">mobile devices and IoT</a> ] \n'
          h+='</i><br>'

          h+='<i>Mispredictions shared by the community:\n'
          h+='[ <a href="http://cknowledge.org/repo/web.php?wcid='+work['self_module_uid']+':'+dlabels+'">images and classifications</a> ] \n'
          h+='</i><br><br>'


          fc=i.get('file_content_base64','')
          fcu=i.get('file_content_uploaded','')

          if dar: 
             fc=''
             fcu=''

          if 'dnn_add_correct_label' in i:
             # Record correct label
             ftmp=i.get('dnn_saved_image_file','')
             label=i.get('dnn_correct_label','').strip()

             wlabel=i.get('dnn_original_classification','').strip()
 
             if label!='' and os.path.isfile(ftmp):
                # Check if already has holder for correct images
                r=ck.access({'action':'find',
                             'module_uoa':work['self_module_uid'],
                             'data_uoa':dlabels,
                             'repo_uoa':ck.cfg.get('record_local_repo_uoa','')})
                if r['return']>0 and r['return']!=16: return r
                if r['return']==16:
                   r=ck.access({'action':'add',
                                'module_uoa':work['self_module_uid'],
                                'data_uoa':dlabels,
                                'repo_uoa':ck.cfg.get('record_local_repo_uoa','')})
                   if r['return']>0: return r
                pl=r['path']

                # Copy image
                pl1=os.path.join(pl,os.path.basename(ftmp))
                shutil.copy(ftmp,pl1)

                # Record label
                r=ck.save_text_file({'text_file':pl1+'.label', 'string':label})
                if r['return']>0: return r

                if wlabel!='':
                   r=ck.save_text_file({'text_file':pl1+'.wrong_label', 'string':wlabel})
                   if r['return']>0: return r

                # Record label
                h+='<b>Thank you for submitting correct label and participating in a creation of a realistic and collective training sets!</b><br><br>\n'
          else:
             if fc=='' and fcu=='':
                # Select image
                h+='Your JPEG image: <input type="file" name="file_content"><br><br>'
                h+='<input type="submit" value="Classify using CK DNN cloud"><br><br>'
             else:
                # Gen tmp file
                rx=ck.gen_tmp_file({'prefix':'tmp-', 'suffix':'.jpg'})
                if rx['return']>0: return rx
                ftmp=rx['file_name']

                if fcu=='':
                   # Save user image
                   rx=ck.convert_upload_string_to_file({'file_content_base64':fc,
                                                        'filename':ftmp})
                   if rx['return']>0: return rx
                else:
                   shutil.copy(fcu,ftmp)

                h+='<input type="hidden" name="dnn_saved_image_file" value="'+ftmp+'">\n'

                # Classify
                ii={'action':'run',
                    'module_uoa':cfg['module_deps']['program'],
                    'data_uoa':cfg['classify_program'][engine],
                    'cmd_key':'classify_ck_ai_api',
                    'quiet':'yes',
                    'generate_rnd_tmp_dir':'yes',
                    'random':'yes',
                    'env':{'CK_AI_API_IMAGE_FILE':ftmp}}
                r=ck.access(ii)
                if r['return']>0:
                   warning='Problem running DNN engine: '+r['error']
                   h+='<b>WARNING:</b> '+warning+'<br><br>'
                else:
                   ch=r.get('characteristics',{})

                   if ch.get('run_success','')!='yes':
                      warning='Problem running DNN engine: '+ch.get('fail_reason','')
                      h+='<b>WARNING:</b> '+warning+'<br><br>'
                   else:
                      te=ch.get('execution_time','')

                      td=r.get('tmp_dir','')
                      deps=r.get('deps',{})

                      k1=engine+'model'
                      k2='lib-'+engine

                      n1=deps.get(k1,{}).get('name','')+' '+deps.get(k1,{}).get('cus',{}).get('package_extra_name','')
                      n2=deps.get(k2,{}).get('name','')+' '+deps.get(k2,{}).get('cus',{}).get('package_extra_name','')

                      # Attempt to read outputs
                      p1=os.path.join(td,'stdout.log')
                      p2=os.path.join(td,'stderr.log')

                      s1=''
                      s2=''

                      r=ck.load_text_file({'text_file':p1})
                      if r['return']==0: 
                         s1=r['string'].strip()

                      r=ck.load_text_file({'text_file':p2})
                      if r['return']==0:
                         s2=r['string']

                      s=s1+'\n'+s2
                      sx=s.split('\n')

                      sy=''
                      started=False
                      top=''
                      for q in sx:
                          q=q.strip()
                          if started:
                             if top=='': top=q
                             if q=='': break
                             sy+=q+'\n'
                          if q.startswith('-----'):
                             started=True

                      prediction=sy

                      s=sy.strip().replace('\n','<br>')

                      # Create table (left - classification, right - stats)
                      h+='<table border="1" cellpadding="10" cellspacing="0">\n'
                      h+=' <tr>\n'
                      h+='  <td valign="top">\n'

                      h+='   <b>Classification output:</b><br><br>\n'
                      h+=    s+'<br><br>\n'

                      h+='   <input type="hidden" name="dnn_original_classification" value="'+prediction.replace('"','\'')+'">\n'

                      h+='   <center>\n'
                      h+='   <b>If classification is wrong, please provide correct label:</b><br><input type="text" name="dnn_correct_label"><br><br>\n'
                      h+='   <button type="submit" name="dnn_add_correct_label">Submit label</button>\n'
                      h+='   ( <a href="'+url0+'&wcid='+work['self_module_uid']+':'+dlabels+'">view shared labels</a> )\n'
                      h+='   </center>\n'

                      h+='  </td>\n'

                      h+='  <td valign="top">\n'
                      h+='    <b>Engine species:</b> '+n2+'<br><br>'
                      h+='    <b>Model variant: </b> '+n1+'<br><br>'

                      h+='    <b>Execution time: </b> '+( '%.1f' % te )+' sec.<br>'

                      h+='  </td>\n'
                      h+=' </tr>\n'
                      h+='</table>\n'

          h+='<br><br>'
          h+='<button type="submit" name="dnn_action_reset">Start again</button>\n'
          h+='<br><br>'

    if sh=='yes':
       h=''
       st=''

    return {'return':0, 'html':h, 'style':st, 'warning':warning, 'prediction':prediction}

##############################################################################
# open dashboard

def dashboard(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    i['action']='browser'
    i['cid']=''
    i['module_uoa']=''
    i['extra_url']='native_action=show&native_module_uoa=model.image.classification'

    return ck.access(i)

##############################################################################
# CK AI JSON API for web (needed to automatically find such function from higher-level CK AI API)

def ask_ai_web(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    return show(i)

##############################################################################
# return json instead of html

def show_json(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    r=show(i)

    if 'html' in r: del(r['html'])
    if 'style' in r: del(r['style'])

    return r

##############################################################################
# user-friendly html view

def html_viewer(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    h='' #'<hr>'

    # Check host URL prefix and default module/action *********************************************
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':cfg['module_deps']['wfe'],
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']
    url0w=rx['url'] #rx['url_without_template']
    template=rx['template']

    purl=i['url_pull']

    muoa=i['module_uoa']
    duoa=i['data_uoa']

    r=ck.access({'action':'load',
                 'module_uoa':muoa,
                 'data_uoa':duoa})
    if r['return']>0: return r

    p=r['path']

    # First find images
    l=[]
    d=os.listdir(p)

    for f in d:
        if f.endswith('.jpg'):
           y={'file':f, 'original_file':f, 'url':purl}

           f1=os.path.join(p,f+'.label')
           f2=os.path.join(p,f+'.wrong_label')

           if os.path.isfile(f1):
              r=ck.load_text_file({'text_file':f1})
              if r['return']>0: return r
              x1=r['string'].strip()
              y['label']=x1

           if os.path.isfile(f2):
              r=ck.load_text_file({'text_file':f2})
              if r['return']>0: return r
              x2=r['string'].strip().replace('\n','<br>')
              y['wrong_label']=x2

           l.append(y)

    # Trying to add visualization from Mobile crowd-benchmarking/crowd-tuning/crowd-learning entries
    r=ck.access({'action':'search',
                 'module_uoa':'experiment.bench.dnn.mobile',
                 'add_meta':'yes'})
    if r['return']==0:
       lst=r['lst']

       for q in lst:
           duid=q['data_uid']
           p=q['path']

           # FGG: temporal hack
           xurl=purl.replace('42b9a1221eb50259:287d4bee982e03c1','experiment.bench.dnn.mobile:'+duid)
           yurl=url0+'wcid=experiment.bench.dnn.mobile:'+duid

           arr=q['meta'].get('all_raw_results',[])

           for y in arr:
               m=y.get('mispredictions',[])
               for z in m:
                   f=z.get('mispredicted_image','')
                   px=os.path.join(p,f)
                   if os.path.isfile(px):

                      ff=f+'.cached.jpg'
                      pxx=os.path.join(p,ff)
                      if not os.path.isfile(pxx):
                         try:
                            import PIL
                            from PIL import Image

                            img=Image.open(px)

                            wx=float(240/float(img.size[0]))
                            h=int((float(img.size[1])*wx))

                            img=img.resize((240, h))
                            img.save(pxx)
                         except Exception as e:
                            pass

                      if os.path.isfile(pxx):
                         x1=z.get('correct_answer','')
                         x2=z.get('misprediction_results','').replace('\n','<br>')

                         y={'file':ff, 'original_file':f, 'entry_url':yurl, 'url':xurl, 'label':x1, 'wrong_label':x2}
                         l.append(y)

    if len(l)==0:
       h+='<br><center><b>No shared mispredictions - community training set is empty!</b></center><br>\n'
    else:
       h+='<center>\n'
       h+='<h2>Collaborative and realistic training set (mispredictions)</h2><br>\n'
       h+='<table border="1" cellpadding="8" cellspacing="0">\n'
       h+=' <tr>\n'
       h+='  <td align="center" valign="top"><b>#</b></td>\n'
       h+='  <td align="center" valign="top"><b>Image</b></td>\n'
       h+='  <td align="center" valign="top"><b>Correct classification</b></td>\n'
       h+='  <td align="center" valign="top"><b>Misclassification</b></td>\n'
       h+=' </tr>\n'

       q=0
       for y in l:
           q+=1
           f=y['file']
           ff=y['original_file']
           url=y['url']
           eurl=y.get('entry_url','')
           px=url+f
           pxx=url+ff
           x1=y['label']
           x2=y['wrong_label']

           z=str(q)
           if eurl!='':
              z='<a href="'+eurl+'">'+z+'</a>'

           h+=' <tr>\n'
           h+='  <td align="center" valign="top">'+z+'</td>\n'
           h+='  <td align="center" valign="top"><a href="'+pxx+'"><img src="'+px+'" width="120"></a></td>\n'
           h+='  <td align="center" valign="top">'+x1+'</td>\n'
           h+='  <td align="center" valign="top">'+x2+'</td>\n'
           h+=' </tr>\n'

       h+='</table>\n'
       h+='</center>\n'

    h+='<BR><center><i>Please, report any illegal/copyrighted content <a href="mailto:admin@cKnowledge.org">here</a> and we will remove it within 48 hours!</i></center><br><br>'

    h+='<hr>\n'

    return {'return':0, 'html':h, 'show_top':'yes'}



#
# Collective Knowledge (dataset)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# Import all files to meta

def import_all_files(i):
    """
    Input:  {
               data_uoa
               (repo_uoa)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    duoa=i['data_uoa']
    ruoa=i.get('repo_uoa','')

    r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duoa,
                 'repo_uoa':ruoa})
    if r['return']>0: return r

    duid=r['data_uid']
    d=r['dict']
    p=r['path']

    if 'dataset_files' not in d: d['dataset_files']=[]
    dfiles=d['dataset_files']

    dirList=os.listdir(p)
    for fn in dirList:
         p1=os.path.join(p, fn)
         if os.path.isfile(p1):
            if fn not in dfiles:
               dfiles.append(fn)

    r=ck.access({'action':'update',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duid,
                 'repo_uoa':ruoa,
                 'dict':d,
                 'substitute':'yes',
                 'sort_keys':'yes'})
    if r['return']>0: return r

    return {'return':0}

##############################################################################
# TBD: generate new data sets to cover unseen behavior
# See https://scholar.google.com/citations?view_op=view_citation&citation_for_view=IwcnpkwAAAAJ:hqOjcs7Dif8C
#     http://arxiv.org/abs/1506.06256

def generate(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    print ('TBD: generate new data sets to cover unseen behavior')

    ck.out('')
    ck.out('Command line: ')
    ck.out('')

    import json
    cmd=json.dumps(i, indent=2)

    ck.out(cmd)

    return {'return':0}

##############################################################################
# TBD: prune data sets to find minimal representative data set covering behavior
# See https://scholar.google.com/citations?view_op=view_citation&citation_for_view=IwcnpkwAAAAJ:hqOjcs7Dif8C
#     http://arxiv.org/abs/1506.06256

def prune(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    print ('prune data sets to find minimal representative data set covering behavior')

    ck.out('')
    ck.out('Command line: ')
    ck.out('')

    import json
    cmd=json.dumps(i, indent=2)

    ck.out(cmd)

    return {'return':0}

##############################################################################
# check size of all data sets and if less than threshold, add tag "small" - 
# needed not to send huge files during collaborative experiments (crowdtuning) via mobile devices

def check_size(i):
    """
    Input:  {
              (repo_uoa)        - repository UOA
              (data_uoa)        - dataset UOA (can be wildcards)

              (limit)           - size limit (to consider small). By default=500000
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              dict         - final dict with key 'features'={...} 
            }

    """

    import os
    import json

    o=i.get('out','')

    sl=i.get('limit','')
    if sl=='': sl=500000
    sl=int(sl)

    muoa=work['self_module_uid']
    duoa=i.get('data_uoa','')
    ruoa=i.get('repo_uoa','')

    rx=ck.access({'action':'search',
                  'repo_uoa':ruoa,
                  'module_uoa':muoa,
                  'data_uoa':duoa})
    if rx['return']>0: return rx

    lst=rx['lst']

    for q in lst:
        muid=q['module_uid']
        ruid=q['repo_uid']
        duid=q['data_uid']
        duoa=q['data_uoa']

        ck.out('Processing '+duoa+' ...')

        ii={'action':'load',
            'module_uoa':muid,
            'repo_uoa':ruid,
            'data_uoa':duid}

        rx=ck.access(ii) 
        if rx['return']>0: return rx

        dd=rx['dict']
        p=rx['path']

        dfiles=dd.get('dataset_files',[])
        tags=dd.get('tags',[])
        sz=0

        for df in dfiles:
            pp=os.path.join(p, df)
            if os.path.isfile(pp):
               sz+=os.path.getsize(pp) 

        x=''
        if sz<sl:
           x=' (SMALL)'

           if 'small' not in tags:
              tags.append('small')
              dd['tags']=tags

              ii['action']='update'
              ii['dict']=dd
              ii['sort_keys']='yes'
              ii['ignore_update']='yes'
              
              rx=ck.access(ii)
              if rx['return']>0: return rx

        ck.out('  Size: '+str(sz)+x)

    return {'return':0}

##############################################################################
# add file to a given dataset

def add_file_to(i):
    """
    Input:  {
              data_uoa   - dataset entry to add file to
              (repo_uoa) - repository of the entry
              file       - file to add
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import shutil
    import os

    o=i.get('out','')

    duoa=i.get('data_uoa','')
    muoa=i.get('module_uoa','')
    ruoa=i.get('repo_uoa','')

    fn=i.get('file','')

    if duoa=='' or fn=='':
       return {'return':1, 'error':'usage - ck add_file_to dataset:{dataset UOA} --file={filename}'}

    # Load entry
    r=ck.access({'action':'load',
                 'module_uoa':muoa,
                 'data_uoa':duoa,
                 'repo_uoa':ruoa})
    if r['return']>0: return r
    p=r['path']
    d=r['dict']

    # Copy file
    pn=os.path.join(p,fn)

    if o=='con':
       ck.out('Copying file '+fn+' to '+pn+' ...')

    shutil.copyfile(fn,pn)

    # Adding to dataset list
    df=d.get('dataset_files',[])
    df.append(fn)
    d['dataset_files']=df

    # Updating entry
    r=ck.access({'action':'update',
                 'module_uoa':muoa,
                 'data_uoa':duoa,
                 'repo_uoa':ruoa,
                 'dict':d,
                 'sort_keys':'yes'})
    if r['return']>0: return r

    return {'return':0}

##############################################################################
# add dataset 

def add(i):
    """
    Input:  {
               (tags) - use tags (string; tags separated by comma)
               (file) - add file
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    import os
    import shutil

    o=i.get('out','')

    duoa=i.get('data_uoa','')

    d=i.get('dict',{})

    # Check tags
    xtags=d.get('tags',[])
    if len(xtags)==0:
       tags=i.get('tags','').strip()
       if tags=='':
          if o=='con':
             rx=ck.inp({'text':'Enter tags for your data set separated by comma (such as image,jpeg): '})
             if rx['return']>0: return rx
             tags=rx['string'].strip()

       xtags=['dataset']
       for t in tags.split(','):
           t1=t.strip()
           if t1!='':
              if t1 not in xtags:
                 xtags.append(t1)

    d['tags']=xtags

    # Check files
    fn=i.get('file','')
    if fn!='':
       fn1=os.path.basename(fn)

       df=d.get('dataset_files',[])
       if fn1 not in df: 
          df.append(fn1)
       d['dataset_files']=df

    # Create entry
    i['dict']=d

    i['common_func']='yes'
    i['sort_keys']='yes'

    r=ck.access(i)
    if r['return']>0: return r
    p=r['path']

    # Copy file
    if fn!='':
       pn=os.path.join(p,fn1)

       if o=='con':
          ck.out('')
          ck.out('Copying file '+fn+' to '+pn+' ...')

       shutil.copyfile(fn,pn)

    return r



#
# Collective Knowledge (CK)
#
# See CK LICENSE.txt for licensing details.
# See CK COPYRIGHT.txt for copyright details.
#
# Developer: Grigori Fursin
#

# CK dummy

from ck.kernel import init, err

# Initialize various vars and paths
r = init({})
if r['return'] > 0:
    err(r)



#
# Collective Knowledge (web-based user auth)
#
# 
# 
#
# Developer: 
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# authentication of a user

def auth(i):
    """
    Input:  {
              (web_vars_post)
              (web_vars_get)
              (web_vars_session)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              user_id
              session_id
            }

    """

    import re
    import hashlib

    rr={'return':0}

    wvp=i.get('web_vars_post',{})

    # Check username
    username=wvp.get('username','').strip()
    if username=='':
       return {'return':1, 'error':'username is empty'}

    if not re.match("^[A-Za-z0-9.@_-]*$", username):
       return {'return':1, 'error':'username contains forbidden characters'}

    # Check password
    password=wvp.get('password','').strip()
    if password=='':
       return {'return':1, 'error':'password is empty'}

    password_md5=hashlib.md5(password.encode('utf8')).hexdigest()

    # Check if entry exists
    default_repo=ck.cfg.get('default_repo_to_write','')
    if default_repo=='': default_repo='local'

    r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'repo_uoa':default_repo,
                 'data_uoa':username})
    if r['return']>0: 
       if r['return']==16: return {'return':16, 'error':'Username not registered'}
       return r

    d=r['dict']

    load_password_md5=d.get('password_md5','')
    if password_md5!=load_password_md5:
       return {'return':1, 'error':'password did not match'}

    # Generate random token for the session
    r=ck.gen_uid({})
    if r['return']>0: return r

    rr['session_id']=r['data_uid']
    rr['user_id']=username

    return rr

##############################################################################
# create account

def create(i):
    """
    Input:  {
              (web_vars_post)
              (web_vars_get)
              (web_vars_session)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              user_id
              session_id
            }

    """

    import re
    import hashlib

    rr={'return':0}

    wvp=i.get('web_vars_post',{})

    # Check username
    username=wvp.get('username','').strip()
    if username=='':
       return {'return':1, 'error':'username is empty'}

    if not re.match("^[A-Za-z0-9.@_-]*$", username):
       return {'return':1, 'error':'username contains forbidden characters'}

    # Check password
    password=wvp.get('password','').strip()
    if password=='':
       return {'return':1, 'error':'password is empty'}

    password_md5=hashlib.md5(password.encode('utf8')).hexdigest()

    # Check email
    email=wvp.get('email','').strip()
    if email=='':
       return {'return':1, 'error':'email is empty'}

    realname=wvp.get('realname','').strip()

    # Check if entry exists
    default_repo=ck.cfg.get('default_repo_to_write','')
    if default_repo=='': default_repo='local'

    r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'repo_uoa':default_repo,
                 'data_uoa':username})
    if r['return']==0:
       return {'return':32, 'error':'Username already registered'}

    if r['return']!=16: return r

    # Create entry
    d={'password_md5':password_md5,
       'email':email,
       'realname':realname}

    r=ck.access({'action':'add',
                 'module_uoa':work['self_module_uid'],
                 'repo_uoa':default_repo,
                 'data_uoa':username,
                 'dict':d,
                 'sort_key':'yes'})
    if r['return']>0: return r

    # Generate random token for the session
    r=ck.gen_uid({})
    if r['return']>0: return r

    rr['session_id']=r['data_uid']
    rr['user_id']=username

    return rr

##############################################################################
# renew user

def renew(i):
    """
    Input:  {
              (web_vars_post)
              (web_vars_get)
              (web_vars_session)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              user_id
              session_id
            }

    """

    import re
    import hashlib

    rr={'return':0}

    wvp=i.get('web_vars_post',{})

    # Check username
    username=wvp.get('username','').strip()
    if username=='':
       return {'return':1, 'error':'username is empty'}

    if not re.match("^[A-Za-z0-9.@_-]*$", username):
       return {'return':1, 'error':'username contains forbidden characters'}

    # Load user
    default_repo=ck.cfg.get('default_repo_to_write','')
    if default_repo=='': default_repo='local'

    r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'repo_uoa':default_repo,
                 'data_uoa':username})
    if r['return']>0: return r

    d=r['dict']

    # Check password
    password=wvp.get('password','').strip()
    if password!='':
       password_md5=hashlib.md5(password.encode('utf8')).hexdigest()
       d['password_md5']=password_md5

    # Check email
    email=wvp.get('email','').strip()
    if email!='':
       d['email']=email

    realname=wvp.get('realname','').strip()
    if realname!='':
       d['realname']=realname

    # Update entry
    r=ck.access({'action':'update',
                 'module_uoa':work['self_module_uid'],
                 'repo_uoa':default_repo,
                 'data_uoa':username,
                 'dict':d,
                 'substitute':'yes',
                 'sort_key':'yes'})
    if r['return']>0: return r

    return {'return':0}



#
# Collective Knowledge (detect frontier for multi-objective optimizations (such as execution time vs energy vs code size vs faults vs price ...))
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# Filter frontier (leave only best points) - my own greedy and probably not very optimal algorithm
#
# TBD: should leave only a few points otherwise can be quickly too many 
#  particularly if more than 2 dimensions (performance, energy, size, faults)
#
# Note, that we minimize all dimensions. Otherwise, provide reversed dimension.
#
# HELP IS APPRECIATED!

def filter(i):
    """
    Input:  {
              points          - dict with points, each has dict with optimization dimensions (should have the same names)
                              
              (frontier_keys) - list of keys to leave only best points during multi-objective autotuning
                                 (multi-objective optimization)
                                If omited, use all keys

              (reverse_keys)  - list of values associated with above keys. If True, reverse sorting for a give key
                                (by default descending) - can't be used without "frontier_keys" due to lack of order in python dicts2

              (margins)        - list of margins when comparing values, i.e. Vold/Vnew < this number (such as 1.10 instead of 1).
                                 will be used if !=None  
            }

    Output: {
              return         - return code =  0, if successful
                                           >  0, if error
              (error)        - error text if return > 0

              points         - filtered points!
              deleted_points - deleted points
            }

    """

    oo=i.get('out','')

    points=i['points']
    lp=len(points)

    dpoints={}

    uids=list(points.keys())

    if oo=='con':
       ck.out('Original number of points:        '+str(lp))

    fk=i.get('frontier_keys',[])
    fkr=i.get('reverse_keys',[])
    lrk=len(fkr)

    mar=i.get('margins',[])
    lmar=len(mar)

    if lp>1:
       for l0 in range(0,lp,1):
           ul0=uids[l0]
           if ul0!='':
              p0=points[ul0]
 
              # Check if there is at least one point with all better dimensions

              keep=True

              for l1 in range(0,lp,1):
                  ul1=uids[l1]
                  if ul1!='' and ul1!=ul0:
                     p1=points[ul1]

                     better=True

                     if len(fk)>0:
                        ks=fk
                     else:
                        ks=list(p0.keys())

                     for dim in range(0, len(ks)):
                         d0=ks[dim]

                         v0=p0[d0]
                         if v0!=None and v0!='':
                            v0=float(v0)

                            v1=p1.get(d0,None)
                            if v1!=None and v1!='':
                               v1=float(v1)

                               if v1==0: v1=v0/10
                               if v1==0: v1=0.01

                               m=1.0
                               if dim<lmar and mar[dim]!=None: 
                                  m=mar[dim]

                               if dim<lrk and fkr[dim]==True:
                                  if v1==0 or (v0/v1)>m:
                                     better=False
                                     break
                               elif v0==0 or (v0/v1)<m:
                                  better=False
                                  break

                     if better:
                        keep=False
                        break

              if not keep:
                 dpoints[ul0]=points[ul0]
                 del(points[ul0])
                 uids[l0]=''

    lp=len(points)
    if oo=='con':
       ck.out('Number of points after filtering: '+str(lp))

    return {'return':0, 'points':points, 'deleted_points':dpoints}



#
# Collective Knowledge (platform - Neural Network accelerator)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# Detect Neural Network Accelerator

def detect(i):
    """
    Input:  {
              (target)               - if specified, use info from 'machine' module

              (host_os)              - host OS (detect, if omitted)
              (os) or (target_os)    - OS module to check (if omitted, analyze host)

              (device_id)            - device id if remote (such as adb)
              (skip_device_init)     - if 'yes', do not initialize device
              (print_device_info)    - if 'yes', print extra device info

              (skip_info_collection) - if 'yes', do not collect info (particularly for remote)

              (skip_print_os_info)   - if 'yes', do not print OS info

              (exchange)             - if 'yes', exchange info with some repo (by default, remote-ck)
              (share)                - the same as 'exchange'
              (exchange_repo)        - which repo to record/update info (remote-ck by default)
              (exchange_subrepo)     - if remote, remote repo UOA
              (exchange_locally)     - if 'yes', exchange locally

              (extra_info)           - extra info about author, etc (see add from CK kernel)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              features = {
                nn          - Neural Network Accelerator features (properties), unified
                nn_misc     - assorted Neural Network Accelerator features (properties), platform dependent
              }
            }

    """

    import os

    o=i.get('out','')

    oo=''
    if o=='con': oo=o

    # Check if target
    if i.get('target','')!='':
       r=ck.access({'action':'init',
                    'module_uoa':cfg['module_deps']['machine'],
                    'input':i})
       if r['return']>0: return r

    # Various params
    hos=i.get('host_os','')
    tos=i.get('target_os','')
    if tos=='': tos=i.get('os','')
    tdid=i.get('device_id','')

    sic=i.get('skip_info_collection','')
    sdi=i.get('skip_device_init','')
    pdv=i.get('print_device_info','')

    ex=i.get('exchange','')
    if ex=='': ex=i.get('share','')

    einf=i.get('extra_info','')
    if einf=='': einf={}

    # Get OS info
    import copy
    ii=copy.deepcopy(i)
    ii['out']=oo
    if i.get('skip_print_os_info','')=='yes': ii['out']=''
    ii['action']='detect'
    ii['module_uoa']=cfg['module_deps']['platform.cpu']
    rr=ck.access(ii) # DO NOT USE rr further - will be reused as return !
    if rr['return']>0: return rr

    hos=rr['host_os_uid']
    hosx=rr['host_os_uoa']
    hosd=rr['host_os_dict']

    tos=rr['os_uid']
    tosx=rr['os_uoa']
    tosd=rr['os_dict']

    tbits=tosd.get('bits','')

    tdid=rr['device_id']

    # Some params
    ro=tosd.get('redirect_stdout','')
    remote=tosd.get('remote','')
    win=tosd.get('windows_base','')

    stdirs=tosd.get('dir_sep','')

    dv=''
    if tdid!='': dv=' -s '+tdid

    return {'return':1, 'error':'detection is not yet supported'}

##############################################################################
# set frequency

def set_freq(i):
    """
    Input:  {
              (host_os)              - host OS (detect, if omitted)
              (os) or (target_os)    - OS module to check (if omitted, analyze host)

              (device_id)            - device id if remote (such as adb)

              (value) = "max" (default)
                        "min"
                        int value
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    o=i.get('out','')
    oo=''
    if o=='con': oo=o

    v=i.get('value','')
    if v=='': v='max'

    # Various params
    hos=i.get('host_os','')
    tos=i.get('target_os','')
    if tos=='': tos=i.get('os','')
    tdid=i.get('device_id','')

    # Get OS info
    import copy
    ii=copy.deepcopy(i)
    ii['out']=''
    ii['action']='detect'
    ii['module_uoa']=cfg['module_deps']['platform.os']
    ii['skip_info_collection']='yes'
    ii['skip_device_init']='yes'
    rr=ck.access(ii)
    if rr['return']>0: return rr

    hos=rr['host_os_uid']
    hosx=rr['host_os_uoa']
    hosd=rr['host_os_dict']

    tos=rr['os_uid']
    tosx=rr['os_uoa']
    tosd=rr['os_dict']

    tbits=tosd.get('bits','')

    tdid=rr['device_id']

    dir_sep=tosd.get('dir_sep','')

    remote=tosd.get('remote','')

    return {'return':1, 'error':'under construction ...'}

##############################################################################
# viewing entries as html

def show(i):
    """
    Input:  {
              data_uoa
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              html         - generated HTML
            }

    """


    h='<h2>Neural Network Accelerators participating in crowd-tuning</h2>\n'

    h+='<i>Reuse/extend <a href="https://github.com/ctuning/ck-crowdtuning-platforms/tree/master/platform.nn">CK JSON meta information</a> of these neural network accelerators using "ck pull repo:ck-crowdtuning-platforms" ...</i><br><br>\n'

    h+='<table class="ck_table" border="0" cellpadding="6" cellspacing="0">\n'

    # Check host URL prefix and default module/action
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']

    h+=' <tr style="background-color:#cfcfff;">\n'
    h+='  <td><b>\n'
    h+='   #\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Vendor\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Name\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Notes\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   <a href="'+url0+'wcid='+work['self_module_uoa']+':">CK UID</a>\n'
    h+='  </b></td>\n'
    h+=' </tr>\n'

    ruoa=i.get('repo_uoa','')
    muoa=work['self_module_uoa']
    duoa=i.get('data_uoa','')

    r=ck.access({'action':'search',
                 'module_uoa':muoa,
                 'data_uoa':duoa,
                 'repo_uoa':ruoa,
                 'add_info':'yes',
                 'add_meta':'yes'})
    if r['return']>0: 
       return {'return':0, 'html':'Error: '+r['error']}

    lst=r['lst']

    num=0
    for q in sorted(lst, key = lambda x: (x.get('meta',{}).get('features',{}).get('vendor','').upper(), \
                                          x.get('meta',{}).get('features',{}).get('name','').upper())):

        num+=1

        duoa=q['data_uoa']
        duid=q['data_uid']

        meta=q['meta']
        ft=meta.get('features',{})

        ecid=meta.get('extra_cid','')
        notes=meta.get('notes','')

        if ecid!='':
           if notes!='':
              notes+='\n<br><br>\n'
           notes+='<a href="'+url0+'wcid='+ecid+'">Extra description</a>\n'

        vendor=ft.get('vendor','')
        name=ft.get('name','')

        h+=' <tr>\n'
        h+='  <td valign="top">\n'
        h+='   '+str(num)+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+vendor+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+name+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+notes+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   <a href="'+url0+'wcid='+work['self_module_uoa']+':'+duid+'">'+duid+'</a>\n'
        h+='  </td>\n'
        h+=' </tr>\n'

    h+='</table><br><br>\n'

    return {'return':0, 'html':h}

##############################################################################
# browse platform.nn participated in experiment crowdsourcing (crowd-benchmarking and crowd-tuning)

def browse(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    # TBD: should calculate url
    url='https://github.com/ctuning/ck-crowdtuning-platforms/tree/master/platform.nn'

    import webbrowser
    webbrowser.open(url)

    import time
    time.sleep(3)

    url='http://cknowledge.org/repo/web.php?action=index&module_uoa=wfe&native_action=show&native_module_uoa=platform.nn'

    import webbrowser
    webbrowser.open(url)

    return {'return':0}



#
# Collective Knowledge (deal with choices)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# Make next multi-dimensional choice (with state)

def make(i):
    """
    Input:  {
              choices_desc       - dict with description of choices (flat format)
              choices_order      - list of list of flat choice vectors to tune [[],[],...] - 
                                   list of list is needed to be able to enable indedepent 
                                   selection of groups of choices. For example, iterate
                                   over all possible data set sizes + random flags per data set
              choices_selection  - list of dicts with types of selection for each above group
              choices_current    - current vector of choices
              (random_module)    - if !=None, random module with seed
              (pipeline)         - if set, update it with current choices
              (custom_explore)   - enforce exploration params from command line

              (all)              - if 'yes', select all
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              choices_current    - list of updated choices
              choices            - dictionary of flat choices and values
              choices_order      - list of flat choices (to know order if need such as for LLVM opt)
              pipeline           - upated pipeline with choices
                                   also choices and choices_order are added to pipeline
              finish             - if True, iterations are over
            }

    """

    from random import Random

    o=i.get('out','')

    my_random=i.get('random_module',None)
    if my_random==None: my_random=Random()

    finish=False

    cdesc=i['choices_desc']
    corder=i['choices_order']
    csel=i['choices_selection']
    ccur=i['choices_current']

    pipeline=i.get('pipeline',{})
    cexp=i.get('custom_explore',{})

    cd=len(corder)

    al=i.get('all','')

    # Init current choices
    if len(ccur)==0:
       for c in range(0, cd):
           cx=corder[c]
           cy=[]
           for k in range(0,len(cx)):
               cy.append('')
           ccur.append(cy)

    update=False
    nupdate=False # if force update next

    for cx in range(cd-1,-1,-1):
        cc=corder[cx]
        dc=ccur[cx]

        t={}
        if al=='yes': 
           tp='pre-selected'
        elif cexp.get('type','')!='':
           tp=cexp['type']
        else:
           t=csel[cx]
           tp=t.get('type','')

        ti=t.get('iterations','')
        top=t.get('omit_probability','')
        if cexp.get('omit_probability','')!='': top=cexp['omit_probability']
        if top=='': top=0.0
        else: top=float(top)

        # Take descriptions first directly from choices_selection.
        # Later will be taken from choice_desc, if exists
        zchoice=t.get('choice',[])
        zprefix=t.get('explore_prefix','')
        zdefault=t.get('default','')
        zcanomit=t.get('can_omit','')

        zestart=t.get('start','')
        if cexp.get('start','')!='': zestart=cexp['start']

        zestop=t.get('stop','')
        if cexp.get('stop','')!='': zestop=cexp['stop']

        zestep=t.get('step','')
        if cexp.get('step','')!='': zestep=cexp['step']

        if tp=='': tp='random'

        ci=t.get('cur_iter','')
        if ci=='': ci=-1

        if cx==(cd-1) or update or nupdate or ci==-1:
           nupdate=False

           ci+=1

           if ti!='' and ci>=ti:
              ci=0
              update=True
           else:
              update=False

           dvsame=''
           xupdate=False
           for c in range(len(cc)-1,-1,-1):
               cn=cc[c]

               qt=cdesc.get(cn,{})

               if zcanomit!='': yco=zcanomit
               else: yco=qt.get('can_omit','')

               if len(zchoice)>0: yhc=zchoice
               else:
                  yhc=qt.get('choice',[])
                  if len(yhc)==0:
                     yhc=qt.get('choices',[])

               if zprefix!='': yep=zprefix
               else: yep=qt.get('explore_prefix','')

               if tp!='': ytp=t.get('subtype','')
               else: ytp=qt.get('type','')

               if zdefault!='': ydefault=zdefault
               else: ydefault=qt.get('default','')

               dcc=dc[c]
               if yep!='' and dcc.startswith(yep):
                  dcc=dcc[len(yep):]
                  if ytp=='float': dcc=float(dcc)
                  else: dcc=int(dcc)

               if zestart!='': yestart=zestart
               else: yestart=qt.get('explore_start','')
               if zestop!='': yestop=zestop
               else: yestop=qt.get('explore_stop','')
               if zestep!='': yestep=zestep
               else: yestep=qt.get('explore_step','')

               if yestart!='':
                  if ytp=='float':
                     r1=float(yestart)
                     r2=float(yestop)
                     rs=float(yestep)
                  else:
                     r1=int(yestart)
                     r2=int(yestop)
                     rs=int(yestep)

                  rx=(r2-r1+1)/rs

               dv=ydefault

               # If exploration, set first
#               if tp=='parallel-loop' or tp=='loop':
               if yestart!='': 
                  dv=r1
               elif len(yhc)>0:
                    dv=yhc[0]

               if tp=='pre-selected': 
                  dv=dcc
               elif ci!=0 or (tp=='random' or tp=='random-with-next'):
                  lcqx=len(yhc)
                  if tp=='random' or tp=='random-with-next':
                     omit=False
                     if yco=='yes':
                        x=my_random.randrange(0, 1000)
                        if x<(1000.0*top):
                           omit=True

                     if omit:
                        dv=''
                     else:
                        if lcqx>0:
                           ln=my_random.randrange(0, lcqx)
                           dv=yhc[ln]
                        elif yestart!='':
                             if (type(rx)==float or type(rx)==int or type(rx)==ck.type_long) and rx>=1:
                                y=my_random.randrange(0,int(rx))
                             else:
                                # alternatively should print inconsistency
                                y=0
                             dv=r1+(y*rs)

                     if tp=='random-with-next':
                        nupdate=True

                  elif tp=='parallel-random': # Change all dimensions at the same time (if explorable)!
                       lcqx=len(yhc)
                       if dvsame=='':
                          if lcqx>0:
                             ln=my_random.randrange(0, lcqx)
                             dvsame=yhc[ln]
                          elif yestart!='':
                               if (type(rx)==float or type(rx)==int or type(rx)==ck.type_long) and rx>=1:
                                  y=my_random.randrange(0,int(rx))
                               else:
                                  # alternatively should print inconsistency
                                  y=0
                               dvsame=r1+(y*rs)
                       dv=dvsame

                  elif tp=='parallel-loop' or tp=='loop' or tp=='loop-with-next' or tp=='parallel-loop-with-next':
                       dv=dcc
                       if (tp=='parallel-loop' or tp=='parallel-loop-with-next') or c==len(cc)-1 or xupdate:
                          if yestart!='':
                             dv=dcc+rs
                             if dv>r2:
                                dv=r1
                                if tp=='loop': xupdate=True
                                else: 
                                   ci=0
                                   update=True
                             else:
                                xupdate=False
                          else: 
                             # Here we process choices instead of ranges)
                             ln=t.get('tmp_choice_position',0)
                             if dv=='':
                                ln=0
                             else:
                                if tp=='loop-with-next' or tp=='parallel-loop-with-next':
                                   ln+=1
                                else:
                                   if dv in yhc:
                                      ln=yhc.index(dv)
                                      ln+=1
                                   else:
                                      ln=0

                             t['tmp_choice_position']=ln

                             if ln<lcqx:
                                dv=yhc[ln]
                                xupdate=False
                             else:
# Next is wrong, but check compatibility with previous cases!
#                                dv=ydefault
                                dv=yhc[0]
                                if tp=='loop': xupdate=True
                                else:
                                   ci=0
                                   update=True

                       if tp=='loop-with-next' or tp=='parallel-loop-with-next':
                          nupdate=True

                  # Machine learning based probabilistic adaptive sampling of multi-dimensional 
                  # design and optimization speaces via external plugin
                  # See our work on Collective Mind (2014/2015)
                  #
                  # NOTE: moved to external customized autotuner plugins (see autotune pipeline --custom_autotuner)
#                  elif tp=='machine-learning-based' or tp=='model-based' or tp=='adaptive' or tp=='plugin-based' or tp=='customized': 

                  else:
                     return {'return':1, 'error':'unknown autotuning type ('+tp+')'}

               if yep!='' and dv!='': dv=yep+str(dv)
               dc[c]=dv

           if xupdate:
              update=True

        t['cur_iter']=ci

    corder1=[]
    ccur1={}

    if update: # means that all loops were updated
       finish=True 
    else:
       if o=='con': 
          ck.out('')
          ck.out('  Vector of flattened and updated choices:')


       ll=0
       prt=[]

       for q in range(0, len(corder)):
           qq=corder[q]
           vq=ccur[q]
           for q1 in range(0, len(qq)):
               qq1=qq[q1]
               vq1=vq[q1]

               corder1.append(qq1)
               ccur1[qq1]=vq1

               if o=='con':
                  if vq1!='':
                     if len(qq1)>ll: ll=len(qq1)
                     prt.append({'k':qq1, 'v':vq1})

               rx=ck.set_by_flat_key({'dict':pipeline, 'key':qq1, 'value':vq1})
               if rx['return']>0: return rx
               pipeline=rx['dict']

           # Flatten choices and values, and add to pipeline
           # Useful if order of choices is important (say opt flags in LLVM)
           # Will be decoded by a given pipeline, if needed 
           pipeline['choices_order']=corder1
#           pipeline['choices']=ccur1   

       if o=='con' and len(prt)>0:
          for q in prt:
              k=q['k']
              v=q['v']

              j=ll-len(k)

              x=' '*j

              ck.out('    '+k+x+' : '+str(v))

    return {'return':0, 'choices_current':ccur, 'choices_order':corder1, 'choices':ccur1, 'pipeline':pipeline, 'finish':finish}

##############################################################################
# select list

def select_list(i):
    """
    Input:  {
              choices      - simple text list of choices
              (skip_enter) - if 'yes', do not select 0 when entering 0
              (desc)       - description for each choices entry (list of the same size as choices)
              (swap_name)  - if 'yes' show desc first and real name in brackets
            }

    Output: {
              return   - return code =  0, if successful
                                     >  0, if error
              (error)  - error text if return > 0
              choice   - selected text
              position - selected position in the choice list
            }

    """

    se=i.get('skip_enter','')

    sn=(i.get('swap_name','')=='yes')

    lst=i.get('choices',[])
    dsc=i.get('desc',[])

    zz={}
    iz=0
    for iz in range(0, len(lst)):
        z=lst[iz]

        zs=str(iz)
        zz[zs]=z

        if iz<len(dsc):
           zd=dsc[iz]
           if zd!='': 
              if sn:
                 z=zd+' ('+z+')'
              else:
                 z+=' ('+zd+')'

        ck.out(zs+') '+z)

        iz+=1

    ck.out('')
    y='Select item'
    if se!='yes': y+=' (or press Enter for 0)'
    y+=': '

    rx=ck.inp({'text':y})
    x=rx['string'].strip()
    if x=='' and se!='yes': x='0' 

    if x not in zz:
       return {'return':1, 'error':'number is not recognized'}

    dduoa=zz[x]

    return {'return':0, 'choice':dduoa, 'position':x}

##############################################################################
# improved version of UOA selector over the version from CK kernel

##############################################################################
# Universal UOA selector

def select_uoa(i):
    """
    Input:  {
              choices      - list from search function
              (skip_enter) - if 'yes', do not select 0 when user presses Enter
              (skip_sort)  - if 'yes', do not sort array
            }

    Output: {
              return  - return code =  0, if successful
                                    >  0, if error
              (error) - error text if return > 0
              choice  - data UOA
            }

    """

    se=i.get('skip_enter','')

    lst=i.get('choices',[])

    if i.get('skip_sort','')!='yes':
       klst=sorted(lst, key=lambda v: v['data_uoa'])
    else:
       klst=lst

    zz={}
    iz=0

    for z1 in klst:
        z=z1['data_uid']
        zu=z1['data_uoa']
        zname=z1.get('info',{}).get('data_name','')

        zs=str(iz)
        zz[zs]=z

        x=z
        if zname!='' and zname!=zu: x=zname+', '+x

        ck.out(zs+') '+zu+' ('+x+')')

        iz+=1

    ck.out('')
    y='Select UOA'
    if se!='yes': y+=' (or press Enter for 0)'
    y+=': '

    rx=ck.inp({'text':y})
    x=rx['string'].strip()
    if x=='' and se!='yes': x='0' 

    if x not in zz:
       return {'return':1, 'error':'number is not recognized'}

    dduoa=zz[x]

    return {'return':0, 'choice':dduoa}



#
# Collective Knowledge: compiler flags crowdtuning (crowdsource autotuning via spare computers such as mobile devices)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
compiler_choices='#choices#compiler_flags#'

line='================================================================'

fsummary='summary.json'
fclassification='classification.json'
fgraph='tmp-reactions-graph.json'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# show results

def html_viewer(i):
    """      
    Input:  {
              TBD
              (interactive_report) - if 'yes' output keys for interactive report
              (minimal)            - if 'yes', minimal html
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              html
              (style)      - styles - useful for plotting JavaScript-based graphs

              (predicted_opt) - if MILEPOST prediction is used
            }

    """

    import os
    global cfg, work

    orig_module_uid=work['self_module_uid']

    features=i.get('features',{})
    xfeatures={}
    if len(features)>0:
       for k in features:
           xfeatures[k[2:]]=features[k]

    mcfg=i.get('module_cfg',{})
    if len(mcfg)>0: cfg=mcfg

    mwork=i.get('module_work',{})
    if len(mwork)>0: work=mwork

    ir=i.get('interactive_report','')
    mn=i.get('minimal','')

    st=''

    if i.get('force_url','')=='':
       url0=i['url_base']
    else:
       url0=i['force_url']

    ap=i.get('all_params',{})

    ruoa=i.get('repo_uoa','')
    eruoa=i.get('experiment_repo_uoa','') # for interactive articles
    if eruoa!='': ruoa=eruoa

    muoa=work['self_module_uoa']
    muid=work['self_module_uid']
    duoa=i.get('data_uoa','')

    ik=cfg['improvements_keys']

    # Load program optimization entry
    rx=ck.access({'action':'load',
                  'module_uoa':cfg['module_deps']['module'],
                  'data_uoa':cfg['module_deps']['program.optimization']})
    if rx['return']>0: return rx
    urld=rx['dict'].get('url_discuss','')

    # Load Entry
    r=ck.access({'action':'load',
                 'repo_uoa':ruoa,
                 'module_uoa':muoa,
                 'data_uoa':duoa})
    if r['return']>0: 
       return {'return':0, 'html':'<b>CK error:</b> '+r['error']+'!'}

    p=r['path']
    d=r['dict']
    duid=r['data_uid']

    # Load program module to get desc keys
    r=ck.access({'action':'load',
                 'module_uoa':cfg['module_deps']['module'],
                 'data_uoa':cfg['replay_desc']['module_uoa']})
    if r['return']>0: return r
    pdesc=r.get('desc',{})
    xxkey=cfg['replay_desc'].get('desc_key','')
    if xxkey!='':
       pdesc=pdesc.get(xxkey,{})

    h='<center>\n'

    h+='\n\n<script language="JavaScript">function copyToClipboard (text) {window.prompt ("Copy to clipboard: Ctrl+C, Enter", text);}</script>\n\n' 

    h+='<H2>Distinct solutions after online classification ('+cfg['desc']+')</H2>\n'
    h+='</center>\n'

    h+='<p>\n'

    cid=muid+':'+duid

    h+='<table border="0" cellpadding="4" cellspacing="0">\n'
    x=muid
    if muoa!=muid: x+=' ('+muoa+')'
    h+='<tr><td><b>Scenario UID</b></td><td>'+x+'</td></tr>\n'
    h+='<tr><td><b>Data UID</b></td><td>'+duid+'</td></tr>\n'
    h+='<tr><td><td></td></tr>\n'

    url5=ck.cfg.get('wiki_data_web','')

    if url5!='' or urld!='':
       if url5!='':
          x='<a href="'+url5+muid+'_'+duid+'">GitHub wiki</a>'
       if urld!='':
          if x!='': x+=', '
          x+='<a href="'+urld+'">Google group</a>' 

       h+='<tr><td><b>Discuss (optimizations to improve compilers,<br><a href="https://arxiv.org/abs/1506.06256">semantic/data set/hardware features<br>to improve predictions</a>, etc):</b></td><td>'+x+'</td></tr>\n'

       h+='<tr><td><td></td></tr>\n'

    urlx=url0+'action=get&cid='+cfg['module_deps']['program.optimization']+':'+duid+'&scenario_module_uoa='+muid+'&out=json'
    urls=url0+'action=pull&common_action=yes&cid='+muid+':'+duid+'&filename=summary.json'
    urlc=url0+'action=pull&common_action=yes&cid='+muid+':'+duid+'&filename=classification.json'

    x=''
    if urls!='':
       x+='[ <a href="'+urls+'">All solutions in JSON</a> ]'
    if urlc!='':
       if x!='': x+=', '
       x+='[ <a href="'+urlc+'">Solutions\' classification in JSON</a> ]'

    if x!='':
       h+='<tr><td><b>Download:</b></td><td>'+x+'</td></tr>\n'

    h+='<tr><td><b>Reproduce all (with reactions):</b></td><td><i>ck replay '+cid+'</i></td></tr>\n'

    h+='<tr><td><td></td></tr>\n'

    pr=cfg.get('prune_results',[])
    mm=d.get('meta',{})
    em=d.get('extra_meta',{})
    obj=mm.get('objective','')

    for k in pr:
        qd=k.get('desc','')
        qi=k.get('id','')
        qr=k.get('ref_uid','')
        qm=k.get('ref_module_uoa','')

        x=mm.get(qi,'')
        if x!='' and qm!='' and qr!='':
           xuid=mm.get(qr,'')
           if xuid!='':
              x='<a href="'+url0+'wcid='+qm+':'+xuid+'">'+x+'</a>'

        h+='<tr><td><b>'+qd+'</b></td><td>'+x+'</td></tr>\n'

    h+='<tr><td><b>Objective</b></td><td>'+obj+'</td></tr>\n'

    h+='<tr><td></td><td></td></tr>\n'

    kk=0
    for kx in range(0, len(ik)):
        k=ik[kx]
        k1=k.replace('$#obj#$',obj)
        ik[kx]=k1

        if pdesc.get(k1,{}).get('desc','')!='':
           k1=pdesc[k1]['desc']

        kk+=1

        h+='<tr><td><b>Improvement key IK'+str(kk)+'</b></td><td>'+k1+'</td></tr>\n'

    ik0=ik[0] # first key to sort

    h+='</table>\n'

    h+='<p>\n'
    h+='<center>\n'

    bgraph={"0":[], "1":[]} # graph with highest improvements

    # Load summary
    sols=[]

    psum=os.path.join(p, fsummary)
    if os.path.isfile(psum):
       rx=ck.load_json_file({'json_file':psum})
       if rx['return']>0: return rx
       sols=rx['dict']

    # Load classification file
    classification={}
    pcl=os.path.join(p, fclassification)
    if os.path.isfile(pcl):
       rx=ck.load_json_file({'json_file':pcl})
       if rx['return']>0: return rx
       classification=rx['dict']

    # If features, update similarity and find min/max
    predicted_opt=''
    if len(features)>0:
       dist_min=None
       dist_max=None

       for q in sols:
           em=q.get('extra_meta',{})
           suid=q['solution_uid']

           dv=em.get('flat',{}).get('##characteristics#compile#joined_compiler_flags#min','')

           cls=classification.get(suid,{})
           if len(cls)>0:
              wl_best=len(cls.get('best',[]))
              wl_worst=len(cls.get('worst',[]))

              wl_best_prog_uoa=''
              wl_best_cmd_key=''
              if wl_best>0:
                 wl_best_prog_uoa=cls['best'][0].get('workload',{}).get('program_uoa','') # for now only for the 1st program
                                                                                          # howelever later should group all programs with best opt!
                 wl_best_cmd_key=cls['best'][0].get('workload',{}).get('cmd_key','')

                 if wl_best_prog_uoa!='' and wl_best_cmd_key!='':
                    # Try to load program static features
                    ra=ck.access({'action':'load',
                                  'module_uoa':cfg['module_deps']['program.static.features'],
                                  'data_uoa':wl_best_prog_uoa})
                    if ra['return']==0:
                       dfeat=ra['dict'].get('features',{}).get('program_static_milepost_features',{})

                       # Load program to get hot kernel (for features)
                       ra=ck.access({'action':'load',
                                     'module_uoa':cfg['module_deps']['program'],
                                     'data_uoa':wl_best_prog_uoa})
                       if ra['return']==0:
                          func=ra['dict'].get('run_cmds',{}).get(wl_best_cmd_key,{}).get('hot_functions',[])

                          if len(func)>0:
                             dft=dfeat.get(func[0]['name'],{})
   #
                            # Calculate similarity
                             r=ck.access({'action':'calculate_similarity',
                                          'module_uoa':cfg['module_deps']['program.static.features'],
                                          'features1':xfeatures,
                                          'features2':dft})
                             if r['return']==0:
                                dist=r['distance']

                                if dist!=None:
                                   cls['distance']=dist

                                   if dist_max==None or dist>dist_max:
                                      dist_max=dist

                                   if dist_min==None or dist<dist_min:
                                      dist_min=dist

    h+='<p>\n'
    h+='$#graph#$\n'
    h+='<p>\n'

    rrr={'return':0} # Preparing output (useful for interactive graphs)

    # List solutions
    if len(sols)==0:
       h+='<h2>No distinct solutions found!</h2>\n'
    else:
       # Check host URL prefix and default module/action
       h+='<table class="ck_table" border="0">\n'

       h+=' <tr style="background-color:#cfcfff;">\n'
       h+='  <td colspan="1"></td>\n'
       if len(features)>0:
          h+='  <td colspan="1" style="background-color:#bfffbf;"></td>\n'
       h+='  <td colspan="1" style="background-color:#bfbfff;"></td>\n'
       h+='  <td colspan="'+str(len(ik))+'" align="center"><b>Improvements (<4% variation)</b></td>\n'
       h+='  <td colspan="2" align="center" style="background-color:#bfbfff;"></td>\n'
       h+='  <td colspan="4"></td>\n'
       h+='  <td colspan="4" align="center" style="background-color:#bfbfff;"><b>Distinct workload for highest improvement</b></td>\n'
       h+='  <td colspan="4"></td>\n'
       h+='  <td colspan="1" align="center" style="background-color:#bfbfff;"></td>\n'
       h+=' </tr>\n'

       h+=' <tr style="background-color:#cfcfff;">\n'
       h+='  <td><b>\n'
       h+='   #\n'
       h+='  </b></td>\n'

       if len(features)>0:
          h+='  <td style="background-color:#bfffbf;"><b>\n'
          h+='   <a href="http://ctuning.org/wiki/index.php/CTools:MilepostGCC:StaticFeatures:MILEPOST_V2.1">MILEPOST features and NN distance (red - most close)</a>\n'
          h+='  </b></td>\n'

       h+='  <td style="background-color:#bfbfff;"><b>\n'
       h+='   Solution UID\n'
       h+='  </b></td>\n'

       for k in range(0, len(ik)):
           h+='  <td align="right"><b>\n'
           h+='   IK'+str(k+1)+'\n'
           h+='  </b></td>\n'

       h+='  <td align="center" style="background-color:#bfbfff;"><b>\n'
       h+='   New distinct optimization choices\n'
       h+='  </b></td>\n'
       h+='  <td align="center" style="background-color:#bfbfff;" align="right"><b>\n'
       h+='   Ref\n'
       h+='  </b></td>\n'

       h+='  <td align="center"><b>\n'
       h+='   Best species\n'
       h+='  </b></td>\n'
       h+='  <td align="center"><b>\n'
       h+='   Worst species\n'
       h+='  </b></td>\n'
       h+='  <td align="center"><b>\n'
       h+='   Touched\n'
       h+='  </b></td>\n'
       h+='  <td align="center"><b>\n'
       h+='   Iters\n'
       h+='  </b></td>\n'
       h+='  <td style="background-color:#bfbfff;"><b>\n'
       h+='   Program\n'
       h+='  </b></td>\n'
       h+='  <td style="background-color:#bfbfff;"><b>\n'
       h+='   CMD\n'
       h+='  </b></td>\n'
       h+='  <td style="background-color:#bfbfff;"><b>\n'
       h+='   Dataset\n'
       h+='  </b></td>\n'
       h+='  <td style="background-color:#bfbfff;"><b>\n'
       h+='   Dataset file\n'
       h+='  </b></td>\n'
       h+='  <td align="right"><b>\n'
       h+='   CPU freq (MHz)\n'
       h+='  </b></td>\n'
       h+='  <td align="right"><b>\n'
       h+='   Cores\n'
       h+='  </b></td>\n'
       h+='  <td><b>\n'
       h+='   Platform\n'
       h+='  </b></td>\n'
       h+='  <td><b>\n'
       h+='   OS\n'
       h+='  </b></td>\n'

       h+='  <td align="center" style="background-color:#bfbfff;">\n'
       h+='   <b>Replay</b>\n'
       h+='  </td>\n'

       h+=' </tr>\n'

       # List
       num=0
       iq=-1
       iq1=0

       res={}
       sres=[]
       ires=0

       em={}

       cls={}

       tbl=[]

       while iq1<len(sols): # already sorted by most "interesting" solutions (such as highest speedups)
           xtbl={}

           xreuse_shared_solution=''
           if iq!=iq1:
              num+=1

              iq+=1
              q=sols[iq]

              em=q.get('extra_meta',{})

              suid=q['solution_uid']

              cls=classification.get(suid,{})

              xcls=cls.get('highest_improvements_workload',{})
              program_uoa=xcls.get('program_uoa','')
              cmd=xcls.get('cmd_key','')
              dataset_uoa=xcls.get('dataset_uoa','')
              dataset_file=xcls.get('dataset_file','')

              wl_best=len(cls.get('best',[]))
              wl_worst=len(cls.get('worst',[]))

              dist=cls.get('distance',None)

              xreuse_shared_solution=cfg['module_deps']['program.optimization']+':'+duid+'-'+muid+'-'+suid

              url_wl=url0+'action=get_workloads&cid='+cfg['module_deps']['program.optimization']+':'+duid+'&scenario_module_uoa='+muid+'&solution_uid='+suid+'&out=json'
              url_wl_best=url_wl+'&key=best'
              url_wl_worst=url_wl+'&key=worst'

              res={}
              ref_res={}
              sres=[]
              ires=0

              # Try to load all solutions
              p1=os.path.join(p, suid)

              try:
                 dirList=os.listdir(p1)
              except Exception as e:
                  None
              else:
                  for fn in dirList:
                      if fn.startswith('ckp-') and fn.endswith('.flat.json'):
                         uid=fn[4:-10]

                         px=os.path.join(p1, fn)
                         rx=ck.load_json_file({'json_file':px})
                         if rx['return']>0: return rx
                         d1=rx['dict']

                         px=os.path.join(p1,'ckp-'+uid+'.features_flat.json')
                         if rx['return']>0: return rx
                         d2=rx['dict']

                         x={'flat':d1, 'features_flat':d2}

                         px=os.path.join(p1, 'ckp-'+uid+'.features.json')
                         rx=ck.load_json_file({'json_file':px})
                         if rx['return']>0: return rx
                         dx=rx['dict']

                         if dx.get('permanent','')=='yes':
                            ref_res==x
                         else:
                            res[uid]=x
                         
                  rr=list(res.keys())
                  sres=sorted(rr, key=lambda v: (float(res[v].get('flat',{}).get(ik0,0.0))), reverse=True)

           rr={}
           if ires<len(sres):
              rr=res.get(sres[ires],{})
              ires+=1

              iterations=q.get('iterations',1)
              touched=q.get('touched',1)

              choices=q['choices']

              ref_sol=q.get('ref_choices',{})
              ref_sol_order=q.get('ref_choices_order',[])

              target_os=choices.get('target_os','')

              speedup=''

              cmd1=''
              cmd2=''

              xtbl['solution_num']=num

              ss='S'+str(num)
              h+=' <tr>\n'
              h+='  <td valign="top" style="background-color:#efefff;">\n'
              if ires<2:
                 h+='   '+ss+'\n'
              h+='  </td>\n'

              flags=rr.get('flat',{}).get('##characteristics#compile#joined_compiler_flags#min','')

              if len(features)>0:

                 if dist!=None and dist_min!=None and dist_max!=None:

                    if dist==dist_min: predicted_opt=flags

                    xdist="%.3f" % dist

                    col='FFFFFF'
                    if dist<=1:
                       col1=int(55+((dist-dist_min)/(1-dist_min))*200)
                       col2=hex(col1)[2:]
                       if (col1<16): col2='0'+col2
                       col='FF'+col2+col2
                    else:
                       col1=int(55+(1-((dist-1)/(dist_max-1)))*200)
                       col2=hex(col1)[2:]
                       if (col1<16): col2='0'+col2
                       col=col2+col2+'FF'

                    h+='  <td valign="top" align="right" style="background-color: #'+col+'">'+xdist+'\n'
                 else:
                    h+='  <td>\n'

                 h+='  </td>\n'

              h+='  <td valign="top">\n'
              if ires<2 and urlx!='':
                 h+='   <a href="'+urlx+'&solution_uid='+suid+'">'+suid+'</a>\n'
              h+='  </td>\n'

              xtbl['solution_uid']=suid

              for k in range(0, len(ik)):
                  h+='  <td valign="top" align="right" style="background-color:#efefff;">\n'

#                  dv=rr.get('flat',{}).get(ik[k],'')

                  dv=''
                  dvw=''
                  points=q.get('points',[])
                  iresx=ires-1
#                  if iresx<len(points):
#                     dv=points[iresx].get('improvements_best',{}).get(ik[k],'')
#                     dvw=points[iresx].get('improvements_worst',{}).get(ik[k],'')

                  # Add to graph (first dimension and first solution)
#                  if k==0 and ires==1:
                  dv=cls.get('highest_improvements',{}).get(ik[k],None)
                  dvw=cls.get('highest_degradations',{}).get(ik[k],None)

                  xtbl['highest_improvement_ik'+str(k+1)]=dv
                  xtbl['highest_degradation_ik'+str(k+1)]=dvw

                  if k==0:
                     bgraph['0'].append([ss,dv])
                     bgraph['1'].append([ss,dvw])

                  y=''
                  if type(dv)==int or type(dv)==ck.type_long:
                     y=str(dv)
                  else:
                     try:
                        y=('%.2f' % dv)
                     except Exception as e: 
                        y=dv
                        pass

                  if dv!='':
                     if dv>1.0:
                        y='<span style="color:#bf0000">'+y+'</span>'
                     elif dv!=0:
                        y='<span style="color:#0000bf">'+y+'</span>'

                  h+=str(y)+'\n'
                  h+='  </td>\n'

              xtbl['best_flags']=flags

              h+='  <td valign="top">\n'
              h+='   '+flags+'\n'
              h+='  </td>\n'

              h+='  <td valign="top" align="center">\n'
              if ires<2:
                 # Ideally should add pipeline description somewhere
                 # to properly recreate flags. However since it is most of the time -Ox
                 # we don't need to make it complex at the moment 

                 ry=rebuild_cmd({'choices':ref_sol,
                                 'choices_order':ref_sol_order,
                                 'choices_desc':{}})
                 if ry['return']>0: return ry
                 ref=ry['cmd']

                 h+='   '+ref+'\n'

                 xtbl['ref_flags']=ref

              h+='   \n'
              h+='  </td>\n'

              h+='  <td valign="top" align="center" style="background-color:#efefff;">\n'
              if ires<2:
                 h+='   <a href="'+url_wl_best+'">'+str(wl_best)+'</a>\n'
                 xtbl['best_species']=wl_best
              h+='  </td>\n'

              h+='  <td valign="top" align="center" style="background-color:#efefff;">\n'
              if ires<2:
                 h+='   <a href="'+url_wl_worst+'">'+str(wl_worst)+'</a>\n'
                 xtbl['worst_species']=wl_worst
              h+='  </td>\n'

              h+='  <td valign="top" align="center" style="background-color:#efefff;">\n'
              if ires<2:
                 h+='   '+str(touched)+'\n'
              h+='  </td>\n'

              h+='  <td valign="top" align="center" style="background-color:#efefff;">\n'
              if ires<2:
                 h+='   '+str(iterations)+'\n'
              h+='  </td>\n'

              h+='  <td valign="top">\n'
              if ires<2:
                 h+='   <a href="'+url0+'wcid=program:'+program_uoa+'">'+program_uoa+'</a>\n'
              h+='  </td>\n'

              h+='  <td valign="top">\n'
              if ires<2:
                 h+='   '+cmd+'\n'
              h+='  </td>\n'

              h+='  <td valign="top">\n'
              if ires<2:
                 h+='   <a href="'+url0+'wcid=dataset:'+dataset_uoa+'">'+dataset_uoa+'</a>\n'
              h+='  </td>\n'

              h+='  <td valign="top">\n'
              if ires<2:
                 h+='   <a href="'+url0+'action=pull&common_func=yes&cid=dataset:'+dataset_uoa+'&filename='+dataset_file+'">'+dataset_file+'</a>\n'
              h+='  </td>\n'

#              h+='  <td valign="top" align="right">\n'
#              if ires<2:
#                 h+='   '+str(em.get('kernel_repetitions',-1))+'\n'
#              h+='  </td>\n'

              h+='  <td valign="top" align="right" style="background-color:#efefff;">\n'
              if ires<2:
                 x=''
                 qq=em.get('cpu_cur_freq',[])
                 for q in qq:
                     xq=qq[q]
                     if x!='': x+=', '
                     x+=str(xq)
                 h+='   '+x+'\n'
              h+='  </td>\n'

              h+='  <td valign="top" align="right" style="background-color:#efefff;">\n'
              if ires<2:
                 qq=em.get('cpu_num_proc',1)
                 h+='   '+str(qq)+'\n'
              h+='  </td>\n'

              h+='  <td valign="top" style="background-color:#efefff;">\n'
              if ires<2:
                 h+='   '+str(em.get('platform_name',''))+'\n'
              h+='  </td>\n'

              h+='  <td valign="top" style="background-color:#efefff;">\n'
              if ires<2:
                 h+='   '+str(em.get('os_name',''))+'\n'
              h+='  </td>\n'

              x='ck replay '+cid+' --solution_uid='+suid
              y=ck.cfg.get('add_extra_to_replay','')
              if y!='': x+=' '+y
              xtbl['replay']=x

              h+='    <td valign="top" align="center"><input type="button" class="ck_small_button" style="height:40px;" onClick="copyToClipboard(\''+x+'\');" value="Copy CMD\nto clipboard"><br>\n'
              h+='                                    <input type="button" class="ck_small_button" style="height:40px;" onClick="copyToClipboard(\'--shared_solution_cid='+xreuse_shared_solution+'\');" value="Copy CID\nto clipboard"></td>\n'

              h+=' </tr>\n'

           else:
              iq1+=1

           if len(xtbl)>0:
              tbl.append(xtbl)

       if ir=='yes':
          rrr['table']=tbl

       h+='</table>\n'

       if predicted_opt!='':
          h+='<br><br><b>Using machine learning to predict optimizations:</b><br><br><i>'+predicted_opt+'</i><br><br>\n'
       h+='<br><a href="http://arxiv.org/abs/1506.06256"><img src="'+url0+'action=pull&common_action=yes&cid='+cfg['module_deps']['module']+':'+orig_module_uid+'&filename=images/image-workflow1.png"></a>\n'

    h+='</center>\n'

    h+='<br><br>\n'

    rx=ck.access({'action':'links',
                  'module_uoa':cfg['module_deps']['program.optimization']})
    if rx['return']>0: return rx
    h+=rx['html']

    # Plot graph
    hg=''
    ftmp=''

    d3_div='ck_interactive'
    if i.get('graph_d3_div','')!='': d3_div=i['graph_d3_div']

    if len(bgraph['0'])>0:
       ii={'action':'plot',
           'module_uoa':cfg['module_deps']['graph'],

           "table":bgraph,

           "h_lines":[1.0],

           "ymin":0,

           "ignore_point_if_none":"yes",

           "plot_type":"d3_2d_bars",

           "display_y_error_bar":"no",

           "title":"Powered by Collective Knowledge",

           "axis_x_desc":"Distinct optimization solutions (highest improvement vs highest degradation)",
           "axis_y_desc":"Max improvement ( IK1 = Ref / Solution )",

           "plot_grid":"yes",

           "d3_div":d3_div,

           "image_width":"900",
           "image_height":"400",

           "wfe_url":url0}

       # Trick to save to file (for interactive/live articles)
       if ir=='yes':
          import copy
          rrr['graph_dict']=copy.deepcopy(ii)

       if ap.get('fgg_save_graph_to_file','')=='yes':
          import copy
          iii=copy.deepcopy(ii)

          iii["substitute_x_with_loop"]="yes"
          iii["plot_type"]="mpl_2d_bars" 
          if 'ymin' in iii: del(iii['ymin'])
          if 'ymax' in iii: del(iii['ymax'])

          # Prepare batch file
          rx=ck.gen_tmp_file({'prefix':'tmp-', 'suffix':'.json'})
          if rx['return']>0: return rx
          ftmp=rx['file_name']

          rx=ck.save_json_to_file({'json_file':ftmp, 'dict':iii, 'sort_keys':'yes'})
          if rx['return']>0: return rx

       r=ck.access(ii)
       if r['return']==0:
          x=r.get('html','')
          if x!='':
             st=r.get('style','')

             hg='<div id="ck_box_with_shadow" style="width:920px;">\n'
             if ftmp!='':
                hg+='<center><b>Note: graph info has been saved to file '+ftmp+' for interactive publications</b></center>'
             hg+=' <div id="'+d3_div+'" style="text-align:center">\n'
             hg+=x+'\n'
             hg+=' </div>\n'
             hg+='</div>\n'

    if mn=='yes':
       h=hg
    else:
       h=h.replace('$#graph#$', hg)

    rrr['html']=h
    rrr['style']=st
    rrr['predicted_opt']=predicted_opt

    return rrr

##############################################################################
# crowdsource these experiments

def crowdsource(i):
    """
    Input:  {
              See 'crowdsource program.optimization'

              (compiler_env_uoa)           - fix compiler environment
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    global cfg, work

    import copy

    mcfg=i.get('module_cfg',{})
    if len(mcfg)>0: 
       cfg=mcfg

    mwork=i.get('module_work',{})
    if len(mwork)>0: work=mwork

    # Setting output
    o=i.get('out','')
    oo=''
    if o=='con': oo='con'

    quiet=i.get('quiet','')

    er=i.get('exchange_repo','')
    if er=='': er=ck.cfg['default_exchange_repo_uoa']
    esr=i.get('exchange_subrepo','')
    if esr=='': esr=ck.cfg['default_exchange_subrepo_uoa']

    if i.get('local','')=='yes': 
       er='local'
       esr=''

    la=i.get('local_autotuning','')

    # Get user 
    user=''

    mcfg={}
    ii={'action':'load',
        'module_uoa':'module',
        'data_uoa':cfg['module_deps']['program.optimization']}
    r=ck.access(ii)
    if r['return']==0:
       mcfg=r['dict']

       dcfg={}
       ii={'action':'load',
           'module_uoa':mcfg['module_deps']['cfg'],
           'data_uoa':mcfg['cfg_uoa']}
       r=ck.access(ii)
       if r['return']>0 and r['return']!=16: return r
       if r['return']!=16:
          dcfg=r['dict']

       user=dcfg.get('user_email','')

    ceuoa=i.get('compiler_env_uoa', '')

    if ceuoa!='':
       rx=ck.access({'action':'load',
                     'module_uoa':cfg['module_deps']['env'],
                     'data_uoa':ceuoa})
       if rx['return']>0: return rx
       ceuoa=rx['data_uid']

    # Initialize local environment for program optimization ***********************************************************
    pi=i.get('platform_info',{})
    if len(pi)==0:
       ii=copy.deepcopy(i)
       ii['action']='initialize'
       ii['module_uoa']=cfg['module_deps']['program.optimization']
       ii['exchange_repo']=er
       ii['exchange_subrepo']=esr
       r=ck.access(ii)
       if r['return']>0: return r

       pi=r['platform_info']
       user=r.get('user','')

    hos=pi['host_os_uoa']
    hosd=pi['host_os_dict']

    tos=pi['os_uoa']
    tosd=pi['os_dict']
    tbits=tosd.get('bits','')

    remote=tosd.get('remote','')

    tdid=pi['device_id']

    program_tags=i.get('program_tags','')
    if program_tags=='' and i.get('local_autotuning','')!='yes' and i.get('data_uoa','')=='':
       program_tags=cfg['program_tags']

    # Check that has minimal dependencies for this scenario ***********************************************************
    sdeps=i.get('dependencies',{}) # useful to preset inside crowd-tuning
    if len(sdeps)==0:
       sdeps=copy.deepcopy(cfg['deps'])
    if len(sdeps)>0:
       if o=='con':
          ck.out(line)
          ck.out('Resolving software dependencies required for this scenario ...')
          ck.out('')

       if ceuoa!='':
          x=sdeps.get('compiler',{})
          if len(x)>0:
             if 'cus' in x: del(x['cus'])
             if 'deps' in x: del(x['deps'])
             x['uoa']=ceuoa
             sdeps['compiler']=x

       ii={'action':'resolve',
           'module_uoa':cfg['module_deps']['env'],
           'host_os':hos,
           'target_os':tos,
           'device_id':tdid,
           'deps':sdeps,
           'add_customize':'yes'}
       if quiet=='yes': 
          ii['random']='yes'
       else:
          ii['out']=oo
       rx=ck.access(ii)
       if rx['return']>0: return rx

       sdeps=rx['deps'] # Update deps (add UOA)

    cpu_name=pi.get('features',{}).get('cpu',{}).get('name','')
    compiler_soft_uoa=sdeps.get('compiler',{}).get('dict',{}).get('soft_uoa','')
    compiler_env=sdeps.get('compiler',{}).get('bat','')
    compiler_tool=sdeps.get('compiler',{}).get('dict',{}).get('env',{}).get('CK_CC','')

    plat_extra={}
    pft=pi.get('features',{})
    for q in pft:
        if q.endswith('_uid'):
           plat_extra[q]=pft[q]
        elif type(pft[q])==dict and pft[q].get('name','')!='':
           plat_extra[q+'_name']=pft[q]['name']

    # Detect real compiler version ***********************************************************
    if o=='con':
       ck.out(line)
       ck.out('Detecting compiler version ...')

    ii={'action':'internal_detect',
        'module_uoa':cfg['module_deps']['soft'],
        'data_uoa':compiler_soft_uoa,
        'host_os':hos,
        'target_os':tos,
        'target_device_id':tdid,
        'env':compiler_env,
        'tool':compiler_tool}
    r=ck.access(ii)
    if r['return']>0: return r

    compiler_version=r['version_str']

    compiler=cfg.get('compiler_name','')+' '+compiler_version

    if o=='con':
       ck.out('')
       ck.out('* Compiler: '+compiler)
       ck.out('* CPU:      '+cpu_name)

    # Start preparing input to run program.optimization
    ii=copy.deepcopy(i)

    ii['action']='run'
    ii['module_uoa']=cfg['module_deps']['program.optimization']

    ii['host_os']=hos
    ii['target_os']=tos
    ii['target_device_id']=tdid
    ii['dependencies']=sdeps

    ii['scenario_cfg']=cfg

    ii['platform_info']=pi

    ii['program_tags']=program_tags

    ii['scenario_module_uoa']=work['self_module_uid']

    ii['experiment_meta']={'cpu_name':cpu_name,
                           'compiler':compiler}

    ii['experiment_meta_extra']=plat_extra

    ii['exchange_repo']=er
    ii['exchange_subrepo']=esr

    ii['user']=user

    # Select sub-scenario ********************************************************************
    from random import randint
    ss=1 # num of scenarios

    sx=randint(1,ss)

    rr={'return':0}

    if sx==1 or la=='yes':
       # **************************************************************** explore random program/dataset
       sdesc='explore random program/cmd/data set'
       if o=='con':
          ck.out('')
          ck.out('  ****** Sub-scenario: '+sdesc+' ******')

       ii['subscenario_desc']=sdesc

       rr=ck.access(ii)
       if rr['return']>0: return rr

    rr['platform_info']=pi

    return rr

##############################################################################
# rebuild compiler cmd from choices

def rebuild_cmd(i):
    """
    Input:  {
               choices       - dict of choices
               choices_order - choices order
               choices_desc  - dict of choices desc
            }

    Output: {
              return         - return code =  0, if successful
                                           >  0, if error
              (error)        - error text if return > 0

              cmd            - compiler command line
              pruned_choices - leave only compiler flags
            }

    """

    cmd=''

    choices=i.get('choices',{})
    corder=i.get('choices_order',[])
    cdesc=i.get('choices_desc',{})

    for q in sorted(corder):
        v=choices.get(q, None)
        d=cdesc.get(q, None)

        if v!=None:
           if cmd!='': cmd+=' '
           cmd+=v

    return {'return':0, 'cmd':cmd}

##############################################################################
# replay optimization

def replay(i):
    """
    See 'replay program.optimization'
    """

    i['module_uoa']=cfg['module_deps']['program.optimization']
    i['module_ref_uoa']=work['self_module_uid']
    i['module_cfg']=copy.deepcopy(cfg)
    i['module_work']=copy.deepcopy(work)
    return ck.access(i)

##############################################################################
# prune compiler flags to find minimal set of choices

def prune(i):
    """
    See 'replay program.optimization'
    """

    i['module_uoa']=cfg['module_deps']['program.optimization']
    i['module_ref_uoa']=work['self_module_uid']
    i['module_cfg']=copy.deepcopy(cfg)
    i['module_work']=copy.deepcopy(work)
    return ck.access(i)

##############################################################################
# prepare graph for interactive reports

def show(i):
    """
    Input:  {
               (from_repo)         - change repository (useful for remote-ck)
               (change_module_uoa) - change module_uoa (to select scenario module)
               (force_url)         - useful to redirect interactive graphs to external repo
               (save_to_file)      - output to file (for auto-generated LaTex and interactive graphs via CK)
               (minimal)           - if 'yes', return minimal html
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    i['action']='process_interactive_graph'
    i['interactive_report']='yes'
    i['out']=''

    from_repo=i.get('from_repo','')
    if from_repo!='':
       i['repo_uoa']=from_repo

    rr=ck.access(i)
    if rr['return']>0: return rr

    if i.get('minimal','')=='yes':
       rr['html']=rr.get('graph_html','')

    stf=i.get('save_to_file','')
    stf0=''
    # Checks to avoid hacking
    if stf!='':
       stf=os.path.basename(stf)
       if not stf.endswith('.png') and not stf.endswith('.pdf'):
          stf=''
       if stf!='':
          stf0=os.path.splitext(stf)[0]

       # Save graph
       iii=rr.get('graph_dict',{})

       iii["substitute_x_with_loop"]="yes"
       iii["plot_type"]="mpl_2d_bars" 

       if 'ymin' in iii: del(iii['ymin'])
       if 'ymax' in iii: del(iii['ymax'])

       iii['out_to_file']=stf

       if i.get("save_graph_title",'')!='': iii['title']=i['save_graph_title']
       if i.get("save_graph_axis_x_desc",'')!='': iii['axis_x_desc']=i['save_graph_axis_x_desc']
       if i.get("save_graph_axis_y_desc",'')!='': iii['axis_y_desc']=i['save_graph_axis_y_desc']
       if i.get("save_graph_axis_x_rotation",'')!='': iii['axis_x_rotation']=i['save_graph_axis_x_rotation']
       if i.get("save_font_size",'')!='': iii['font_size']=i['save_font_size']

       r=ck.access(iii)
       if r['return']>0: return r

       h ='<table border="1" cellpadding="3" cellspacing="0">\n'
       h+=' <tr>\n'
       h+='  <td align="right"><b>Solution</b></td>\n'
       h+='  <td><b>Pruned flags (complexity reduction)</b></td>\n'
       h+='  <td><b>Best species</b></td>\n'
       h+='  <td><b>Worst species</b></td>\n'
       h+=' </tr>\n'

       t ="    \\begin{tabular}{|r|p{4.5in}|p{0.5in}|p{0.5in}|}\n"
       t+="     \\hline\n"
       t+="     \\textbf{Solution} & \\textbf{Pruned flags (complexity reduction)} & \\textbf{Best species} & \\textbf{Worst species} \\\\ \n"
       t+="     \\hline\n"

       tbl=rr.get('table',[])

       for q in tbl:
           sn=q['solution_num']
           bf=q['best_flags']

           bs=q.get('best_species', 0)
           ws=q.get('worst_species', 0)

           t+="      "+str(sn)+" & "+bf+" & "+str(bs)+" & "+str(ws)+" \\\\\n"
           t+="     \\hline\n"

           h+=' <tr>'
           h+='  <td>'+str(sn)+'</td>\n'
           h+='  <td>'+bf+'</td>\n'
           h+='  <td>'+str(bs)+'</td>\n'
           h+='  <td>'+str(ws)+'</td>\n'
           h+=' </tr>'

       t+="    \\end{tabular}"

       h+='</table>\n'

       # Save tex file
       r=ck.save_text_file({'text_file':stf0+'.solutions.tex', 'string':t})
       if r['return']>0: return r

       # Save html file
       r=ck.save_text_file({'text_file':stf0+'.solutions.html', 'string':h})
       if r['return']>0: return r

    return rr

##############################################################################
# prepare graph for interactive reports

def process_interactive_graph(i):
    """
    Input:  {
               (change_module_uoa)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    i['action']='html_viewer'
    i['out']=''

    change_module_uoa=i.get('change_module_uoa','')
    if change_module_uoa!='':
       i['module_uoa']=change_module_uoa

    r=ck.access(i)

    if 'html' in r:
       r['graph_html']=r['html']
       del(r['html'])

    return r



#
# Collective Knowledge (platform - Neural Network accelerator)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# Detect Neural Network Accelerator

def detect(i):
    """
    Input:  {
              (target)               - if specified, use info from 'machine' module

              (host_os)              - host OS (detect, if omitted)
              (os) or (target_os)    - OS module to check (if omitted, analyze host)

              (device_id)            - device id if remote (such as adb)
              (skip_device_init)     - if 'yes', do not initialize device
              (print_device_info)    - if 'yes', print extra device info

              (skip_info_collection) - if 'yes', do not collect info (particularly for remote)

              (skip_print_os_info)   - if 'yes', do not print OS info

              (exchange)             - if 'yes', exchange info with some repo (by default, remote-ck)
              (share)                - the same as 'exchange'
              (exchange_repo)        - which repo to record/update info (remote-ck by default)
              (exchange_subrepo)     - if remote, remote repo UOA
              (exchange_locally)     - if 'yes', exchange locally

              (extra_info)           - extra info about author, etc (see add from CK kernel)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              features = {
                nn          - Neural Network Accelerator features (properties), unified
                nn_misc     - assorted Neural Network Accelerator features (properties), platform dependent
              }
            }

    """
    return {'return':1, 'error':'TBD'}



#
# Collective Knowledge (checking and installing software)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Leo Gordon, leo@dividiti.com
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel)


def init(i):
    """
    Not to be called directly. Sets the path to the vqe_plugin.
    """

    return {'return':0}


def iterate(i):
    """
    Input:  {
                (any_params)        - any command line parameters, some of which (being iterable) given as lists/ranges

                    Please note the expected command line syntax for iterable values:

                --alpha=val         - a single value of any type

                --beta,=10,20,30    - a list of values (comma-separated)
                --beta:=10:20:30    - a list of values (colon-separated)

                --delta:12:18       - a range of integer values (step=1 by default), inclusive of both lower and upper bounds.
                --delta:12:18:3     - a range of integer values (step=3), inclusive of both lower and upper bounds.
            }

    Output: {
                return              - return code =  0, if successful
                                                  >  0, if error
                (error)             - error text if return > 0
            }
    Test:
            ck iterate cmdgen --alpha,=11,22,33 --beta:-9:-7 --delta=single --gamma:=one:two:three
            ck iterate cmdgen --alpha,=11,22,33 --beta:-9:-7 --delta=single --gamma:=one:two:three --out=json
    """
    import re

    interactive     = i.get('out')=='con'

    input_params    = i.copy()
    for unwanted_key in ('action', 'repo_uoa', 'module_uoa', 'data_uoa', 'cid', 'cids', 'xcids', 'out'):
        input_params.pop(unwanted_key, None)

    index_name  = []
    index_range = []

    for param_name in input_params.keys():
        param_value = input_params[param_name]
        matchObj = re.match('(\w+)([,:])?((-?\d+):(-?\d+)(?:\:(\d+))?)?$', param_name)
        if matchObj:
            pure_name   = matchObj.group(1)
            index_name.append( pure_name )
            if matchObj.group(2)==None:
                index_range.append( [param_value] )
            elif matchObj.group(3):
                range_from  = int(matchObj.group(4))
                range_to    = int(matchObj.group(5))
                range_step  = int(matchObj.group(6)) if matchObj.group(6) else 1
                index_range.append( [ str(n) for n in range(range_from, range_to+1, range_step) ] )
            else:
                delimiter   = matchObj.group(2)
                index_range.append( param_value.split(delimiter) )
        else:
            return {'return':1, 'error':"Could not recognize parameter '{}'".format(param_name)}

    if interactive:
        print(dict(zip(index_name, index_range)))
        print('-'*80)

    dimensions  = len(index_name)
    multi_idx   = [0] * dimensions
    current_dim = dimensions-1
    param_dicts = []
    while True:
        multi_value = {index_name[i]: index_range[i][multi_idx[i]] for i in range(dimensions) }
        param_dicts.append( multi_value )

        if interactive:
            print(multi_value)

        if current_dim>=0:
            multi_idx[current_dim] += 1
        # carry avalanche:
        while current_dim>=0 and multi_idx[current_dim]>=len(index_range[current_dim]):
            multi_idx[current_dim] = 0
            current_dim -= 1
            if current_dim>=0:
                multi_idx[current_dim] += 1

        if current_dim>=0:
            current_dim = dimensions-1
        else:
            break

    return {'return':0, 'param_dicts': param_dicts}


def gen(i):
    """
    Input:  {
                data_uoa            - specific entry that contains recipe for building the command
                (any_map_params)    - any command line parameters (supported by build_map)

                    Please note the expected command line syntax for iterable values:

                --alpha=val         - a single value of any type

                --beta,=10,20,30    - a list of values (comma-separated)
                --beta:=10:20:30    - a list of values (colon-separated)

                --delta:12:18       - a range of integer values (step=1 by default), inclusive of both lower and upper bounds.
                --delta:12:18:3     - a range of integer values (step=3), inclusive of both lower and upper bounds.
            }

    Output: {
                return              - return code =  0, if successful
                                                  >  0, if error
                (error)             - error text if return > 0
            }
    Test:
            ck gen cmdgen:ls
            ck gen cmdgen:ls --mode=long --home
            ck gen cmdgen:ls --mode,=long,short,single
            ck gen cmdgen:ls --mode=short --files:=$PATH
    """

    import re
    from copy import deepcopy
    from pprint import pprint

    interactive     = i.get('out')=='con'

    load_adict = {  'action':           'load',
                    'repo_uoa':         i.get('repo_uoa', ''),
                    'module_uoa':       i['module_uoa'],
                    'data_uoa':         i['data_uoa'],
    }
    r=ck.access( load_adict )
    if r['return']>0: return r

    entry_dict  = r['dict']
    if interactive:
        print("Entry contents:")
        pprint(entry_dict)
        print("")

    build_map   = entry_dict.get('build_map', {})
    defaults    = entry_dict.get('defaults', {})

    i['action'] = 'iterate'
    i['out']    = ''
    r=ck.access( i )
    if r['return']>0: return r
    param_dicts = r['param_dicts']

    cmds    = []
    for input_params in param_dicts:

        # Accumulating values:
        #
        accu        = deepcopy( entry_dict.get('accu_init', {}) )
        for param_name in input_params:
            param_value = input_params[param_name]

            if param_name in build_map:
                specific_accu_map = build_map[param_name]

                # Start with the specific value:
                accu_map = specific_accu_map.get(param_value)

                # if not found, try matching a pattern:
                if accu_map == None:
                    for pattern_candidate in specific_accu_map:
                        if re.search('^'+pattern_candidate+'$', param_value):
                            accu_map = specific_accu_map[pattern_candidate]

                # if still not found, try the universal default:
                if accu_map == None:
                    accu_map = specific_accu_map.get('###')

                # if still not found, fail gracefully:
                if accu_map==None:
                    return {'return':1, 'error':"build_map[{}] is missing both '{}' and '###' values".format(param_name, param_value)}

                for accu_name in accu_map:
                    if accu_name not in accu:
                        accu[accu_name] = []    # manual vivification

                    accu_value_list = accu_map[accu_name]
                    if type(accu_value_list)!=list:
                        accu_value_list = [ accu_value_list ]

                    for accu_value in accu_value_list:
                        accu[accu_name].append( accu_value.replace('###', param_value) )

        if False and interactive:
            print('-'*80)
            print("Accu contents:")
            pprint(accu)
            print("")

        # Substitute the accumulated values into command template:
        #
        anchor_regexpr  = '((?:<<<|{{{)(\??)(\w+)(.?)(?:>>>|}}}))'
        subst_output    = entry_dict['cmd_template']
        can_substitute  = bool( re.search(anchor_regexpr, subst_output) )
        iteration       = 0

        while can_substitute:
            subst_input = subst_output
            for match in re.finditer(anchor_regexpr, subst_input):
                expression, optional, anchor_name, accu_sep = match.group(1), match.group(2), match.group(3), match.group(4)
                if anchor_name in accu:
                    if anchor_name in input_params:
                        return {'return':1, 'error':"Both input_params and accu contain '{}' anchor, ambiguous substitution".format(anchor_name)}
                    else:
                        accu_value_list = accu[anchor_name]
                        if type(accu_value_list)!=list:
                            accu_value_list = [ accu_value_list ]
                        subst_output = subst_output.replace(expression, accu_sep.join(accu_value_list) )

                        # print("Substituting {} -> {} from accu".format(expression, accu_sep.join(accu[anchor_name])))
                elif anchor_name in input_params:
                    subst_output = subst_output.replace(expression, input_params[anchor_name] )

                    # print("Substituting {} -> {} from input_params".format(expression, input_params[anchor_name]))
                elif anchor_name in defaults:
                    default_value_list = defaults[anchor_name]
                    if type(default_value_list)!=list:
                        default_value_list = [ default_value_list ]
                    subst_output = subst_output.replace(expression, accu_sep.join(default_value_list) )

                    # print("Substituting {} -> {} from defaults".format(expression, defaults[anchor_name]))
                elif optional=='?':
                    # print("Substituting optional {} -> ''".format(expression))
                    subst_output = subst_output.replace(expression, '')
                else:
                    return {'return':1, 'error':"Neither input_params nor accu contain substitution for non-optional '{}' anchor".format(anchor_name)}

                # print("input={}\noutput={}\n".format(subst_input, subst_output))

            can_substitute  = bool( re.search(anchor_regexpr, subst_output) )
            if interactive:
                iteration += 1
                # print("Substitution iteration #{}".format(iteration))


        if subst_output not in cmds:
            cmds.append( subst_output )
            if interactive:
                print('# ' + '-'*40 +' cmd #{}: '.format(len(cmds)) + '-'*40 )
                print("\n{}\n".format(subst_output.replace(' --', ' \\\n    --')))

    return { 'return': 0, 'cmds': cmds }


def run(i):
    """
    Input:  {
                data_uoa            - specific entry that contains recipe for building the command
                (any_map_params)    - any command line parameters (supported by build_map)
            }

    Output: {
                return              - return code =  0, if successful
                                                  >  0, if error
                (error)             - error text if return > 0
            }
    Test:
            ck run cmdgen:ls
            ck run cmdgen:ls --mode=long --home
            ck run cmdgen:ls --mode,=long,short,single
            ck run cmdgen:ls --mode=short --files:=$PATH
    """

    import os

    interactive = i.get('out')=='con'

    i['action'] = 'gen'
    i['out']    = ''
    r=ck.access( i )
    if r['return']>0: return r
    cmds = r['cmds']

    for cmd_idx in range(len(cmds)):
        cmd = cmds[cmd_idx]
        if interactive:
            print('# ' + '='*40 +' cmd #{}: '.format(cmd_idx+1) + '='*40 )
            print("Running command:\n\t"+cmd)
            print('')
        os.system(cmd)

    if interactive:
        print('='*90)

    return { 'return': 0 }


def show(i):
    """
    Input:  {
                (data_uoa)          - name of the SUT entry
            }

    Output: {
                return              - return code =  0, if successful
                                                  >  0, if error
                (error)             - error text if return > 0
            }
    Test:
            ck show sut:velociti
    """

    from pprint import pprint

    interactive     = i.get('out')=='con'
    data_uoa = i.get('data_uoa')
    if data_uoa:
        load_adict = {  'action':           'load',
                        'module_uoa':       i['module_uoa'],
                        'data_uoa':         data_uoa,
        }
        r=ck.access( load_adict )
        if r['return']>0: return r

        entry_dict = r['dict']

        if interactive:
            pprint(entry_dict)

    return { 'return': 0 }


def map_keys(i):
    """
    Input:  {
                data_uoa            - specific entry that contains recipe for building the command
            }

    Output: {
                return              - return code =  0, if successful
                                                  >  0, if error
                (error)             - error text if return > 0
            }
    Test:
            ck map_keys cmdgen:ls
            ck map_keys cmdgen:benchmark_tflite_loadgen
    """

    interactive     = i.get('out')=='con'

    load_adict = {  'action':           'load',
                    'repo_uoa':         i.get('repo_uoa', ''),
                    'module_uoa':       i['module_uoa'],
                    'data_uoa':         i['data_uoa'],
    }
    r=ck.access( load_adict )
    if r['return']>0: return r

    entry_dict  = r['dict']
    build_map   = entry_dict.get('build_map', {})
    map_keys    = list(build_map.keys())

    if interactive:
        print("\n".join(map_keys))

    return { 'return': 0, 'map_keys': map_keys }


def map_values(i):
    """
    Input:  {
                data_uoa            - specific entry that contains recipe for building the command
                option              - the map name for which the values are sought
            }

    Output: {
                return              - return code =  0, if successful
                                                  >  0, if error
                (error)             - error text if return > 0
            }
    Test:
            ck map_values cmdgen:ls --option=mode
    """

    interactive     = i.get('out')=='con'

    load_adict = {  'action':           'load',
                    'repo_uoa':         i.get('repo_uoa', ''),
                    'module_uoa':       i['module_uoa'],
                    'data_uoa':         i['data_uoa'],
    }
    r=ck.access( load_adict )
    if r['return']>0: return r

    entry_dict  = r['dict']
    build_map   = entry_dict.get('build_map', {})
    option      = i['option']
    map_values  = list(build_map[option].keys())

    if interactive:
        print("\n".join(map_values))

    return { 'return': 0, 'map_values': map_values }

